@@ -1,29 +1,63 @@
 
 import argparse
+import csv
 import hashlib
 import logging
 import multiprocessing as mp
 import os
 import sys
 import subprocess as sp
+import re
+import collections
 
 import bakta
 import bakta.constants as bc
 import bakta.config as cfg
 
+
 log = logging.getLogger('UTILS')
 
 
-def parse_arguments():
+def print_version(self):
+    return f'v{self.major}.{self.minor}.{self.patch}'
+
+Version = collections.namedtuple('Version', ['major', 'minor', 'patch'], defaults=[0, 0]) # named tuple for version checking, defaults are zero for missing minor/patch
+Version.__str__ = print_version
+
+
+VERSION_MIN_DIGIT = -1
+VERSION_MAX_DIGIT = 1000000000000
+VERSION_REGEX = re.compile(r'(\d+)\.(\d+)(?:\.(\d+))?')  # regex to search for version number in tool output. Takes missing patch version into consideration.
+
+# List of dependencies: tuples for: min version, max version, tool name & command line parameter, dependency check exclusion options
+DEPENDENCY_TRNASCAN = (Version(2,0,6), Version(VERSION_MAX_DIGIT, VERSION_MAX_DIGIT, VERSION_MAX_DIGIT), VERSION_REGEX, ('tRNAscan-SE', '-h'), ['--skip-trna'])
+DEPENDENCY_ARAGORN = (Version(1,2,38), Version(VERSION_MAX_DIGIT, VERSION_MAX_DIGIT, VERSION_MAX_DIGIT), VERSION_REGEX, ('aragorn', '-h'), ['skip-tmrna'])
+DEPENDENCY_CMSCAN = (Version(1,1,2), Version(VERSION_MAX_DIGIT, VERSION_MAX_DIGIT, VERSION_MAX_DIGIT), VERSION_REGEX, ('cmscan', '-h'), ['--skip-rrna', '--skip-ncrna', '--skip-ncrna-region'])
+DEPENDENCY_PILERCR = (Version(1,6), Version(VERSION_MAX_DIGIT, VERSION_MAX_DIGIT, VERSION_MAX_DIGIT), VERSION_REGEX, ('pilercr', '-options'), ['--skip-crispr'])
+DEPENDENCY_PRODIGAL = (Version(2,6,3), Version(VERSION_MAX_DIGIT, VERSION_MAX_DIGIT, VERSION_MAX_DIGIT), VERSION_REGEX, ('prodigal', '-v'), ['--skip-cds'])
+DEPENDENCY_HMMSEARCH = (Version(3,3,1), Version(VERSION_MAX_DIGIT, VERSION_MAX_DIGIT, VERSION_MAX_DIGIT), VERSION_REGEX, ('hmmsearch', '-h'), ['--skip-cds', '--skip-sorf'])
+DEPENDENCY_DIAMOND = (Version(2,0,4), Version(VERSION_MAX_DIGIT, VERSION_MAX_DIGIT, VERSION_MAX_DIGIT), VERSION_REGEX, ('diamond', 'help'), ['--skip-cds', '--skip-sorf'])
+DEPENDENCY_BLASTN = (Version(2,7,1), Version(VERSION_MAX_DIGIT, VERSION_MAX_DIGIT, VERSION_MAX_DIGIT), VERSION_REGEX, ('blastn', '-version'), ['--skip-ori'])
+DEPENDENCY_AMRFINDERPLUS = (Version(3,10), Version(VERSION_MAX_DIGIT, VERSION_MAX_DIGIT, VERSION_MAX_DIGIT), VERSION_REGEX, ('amrfinder', '--version'), ['--skip-cds'])
+
+
+def init_parser():
     parser = argparse.ArgumentParser(
         prog='bakta',
         description='Rapid & standardized annotation of bacterial genomes & plasmids.',
+        epilog=f'Citation:\n{bc.CITATION}',
+        formatter_class=argparse.RawDescriptionHelpFormatter,
         add_help=False
     )
-    parser.add_argument('genome', metavar='<genome>', help='(Draft) genome in fasta format')
+    return parser
+
+
+def parse_arguments():
+    parser = init_parser()
+    parser.add_argument('genome', metavar='<genome>', help='Genome sequences in (zipped) fasta format')
 
     arg_group_io = parser.add_argument_group('Input / Output')
-    arg_group_io.add_argument('--db', '-d', action='store', default=None, help='Database path (default = <bakta_path>/db)')
+    arg_group_io.add_argument('--db', '-d', action='store', default=None, help='Database path (default = <bakta_path>/db). Can also be provided as BAKTA_DB environment variable.')
     arg_group_io.add_argument('--min-contig-length', '-m', action='store', type=int, default=1, dest='min_contig_length', help='Minimum contig size (default = 1)')
     arg_group_io.add_argument('--prefix', '-p', action='store', default=None, help='Prefix for output files')
     arg_group_io.add_argument('--output', '-o', action='store', default=os.getcwd(), help='Output directory (default = current working directory)')
@@ -35,150 +69,142 @@ def parse_arguments():
     arg_group_organism.add_argument('--plasmid', action='store', default=None, help='Plasmid name')
     
     arg_group_annotation = parser.add_argument_group('Annotation')
-    arg_group_annotation.add_argument('--complete', action='store_true', help="All sequences are complete replicons (chromosome/plasmid[s])")
+    arg_group_annotation.add_argument('--complete', action='store_true', help='All sequences are complete replicons (chromosome/plasmid[s])')
     arg_group_annotation.add_argument('--prodigal-tf', action='store', default=None, dest='prodigal_tf', help='Path to existing Prodigal training file to use for CDS prediction')
-    arg_group_annotation.add_argument('--translation-table', action='store', type=int, default=11, choices=[11, 4], dest='translation_table', help='Translation table to use: 11/4 (default = 11)')
+    arg_group_annotation.add_argument('--translation-table', action='store', type=int, default=11, choices=[11, 4], dest='translation_table', help='Translation table: 11/4 (default = 11)')
     arg_group_annotation.add_argument('--gram', action='store', default='?', choices=['+', '-', '?'], help="Gram type: +/-/? (default = '?')")
-    arg_group_annotation.add_argument('--locus', action='store', default=None, help="Locus prefix (instead of 'contig')")
-    arg_group_annotation.add_argument('--locus-tag', action='store', default=None, dest='locus_tag', help='Locus tag prefix')
+    arg_group_annotation.add_argument('--locus', action='store', default=None, help="Locus prefix (default = 'contig')")
+    arg_group_annotation.add_argument('--locus-tag', action='store', default=None, dest='locus_tag', help='Locus tag prefix (default = autogenerated)')
     arg_group_annotation.add_argument('--keep-contig-headers', action='store_true', dest='keep_contig_headers', help='Keep original contig headers')
-    arg_group_annotation.add_argument('--replicons', '-r', action='store', default=None, dest='replicons', help="Replicon information table (TSV)")
+    arg_group_annotation.add_argument('--replicons', '-r', action='store', default=None, dest='replicons', help='Replicon information table (tsv/csv)')
 
     arg_group_workflow = parser.add_argument_group('Workflow')
-    arg_group_workflow.add_argument('--skip-trna', action='store_true', dest='skip_trna', help="Skip tRNA detection & annotation")
-    arg_group_workflow.add_argument('--skip-tmrna', action='store_true', dest='skip_tmrna', help="Skip tmRNA detection & annotation")
-    arg_group_workflow.add_argument('--skip-rrna', action='store_true', dest='skip_rrna', help="Skip rRNA detection & annotation")
-    arg_group_workflow.add_argument('--skip-ncrna', action='store_true', dest='skip_ncrna', help="Skip ncRNA detection & annotation")
-    arg_group_workflow.add_argument('--skip-ncrna-region', action='store_true', dest='skip_ncrna_region', help="Skip ncRNA region detection & annotation")
-    arg_group_workflow.add_argument('--skip-crispr', action='store_true', dest='skip_crispr', help="Skip CRISPR array detection & annotation")
-    arg_group_workflow.add_argument('--skip-cds', action='store_true', dest='skip_cds', help="Skip CDS detection & annotation")
-    arg_group_workflow.add_argument('--skip-sorf', action='store_true', dest='skip_sorf', help="Skip sORF detection & annotation")
-    arg_group_workflow.add_argument('--skip-gap', action='store_true', dest='skip_gap', help="Skip gap detection & annotation")
-    arg_group_workflow.add_argument('--skip-ori', action='store_true', dest='skip_ori', help="Skip oriC/oriT detection & annotation")
+    arg_group_workflow.add_argument('--skip-trna', action='store_true', dest='skip_trna', help='Skip tRNA detection & annotation')
+    arg_group_workflow.add_argument('--skip-tmrna', action='store_true', dest='skip_tmrna', help='Skip tmRNA detection & annotation')
+    arg_group_workflow.add_argument('--skip-rrna', action='store_true', dest='skip_rrna', help='Skip rRNA detection & annotation')
+    arg_group_workflow.add_argument('--skip-ncrna', action='store_true', dest='skip_ncrna', help='Skip ncRNA detection & annotation')
+    arg_group_workflow.add_argument('--skip-ncrna-region', action='store_true', dest='skip_ncrna_region', help='Skip ncRNA region detection & annotation')
+    arg_group_workflow.add_argument('--skip-crispr', action='store_true', dest='skip_crispr', help='Skip CRISPR array detection & annotation')
+    arg_group_workflow.add_argument('--skip-cds', action='store_true', dest='skip_cds', help='Skip CDS detection & annotation')
+    arg_group_workflow.add_argument('--skip-sorf', action='store_true', dest='skip_sorf', help='Skip sORF detection & annotation')
+    arg_group_workflow.add_argument('--skip-gap', action='store_true', dest='skip_gap', help='Skip gap detection & annotation')
+    arg_group_workflow.add_argument('--skip-ori', action='store_true', dest='skip_ori', help='Skip oriC/oriT detection & annotation')
 
     arg_group_general = parser.add_argument_group('General')
     arg_group_general.add_argument('--help', '-h', action='help', help='Show this help message and exit')
     arg_group_general.add_argument('--verbose', '-v', action='store_true', help='Print verbose information')
     arg_group_general.add_argument('--threads', '-t', action='store', type=int, default=mp.cpu_count(), help='Number of threads to use (default = number of available CPUs)')
     arg_group_general.add_argument('--tmp-dir', action='store', default=None, dest='tmp_dir', help='Location for temporary files (default = system dependent auto detection)')
-    arg_group_general.add_argument('--version', action='version', version='%(prog)s ' + bakta.__version__)
-    arg_group_general.add_argument('--citation', action='store_true', help='Print citation')
+    arg_group_general.add_argument('--version', action='version', version=f'%(prog)s {bakta.__version__}')
     return parser.parse_args()
 
 
-def test_dependencies():
-    """Test the proper installation of necessary 3rd party executables."""
-
-    # test tRNAscan-SE
-    if(cfg.skip_trna is False):
+def read_tool_output(dependency):
+        """Method for reading tool version with regex. Input: regex expression, tool command. Retursn: version number."""
+        version_regex = dependency[2]
+        command = dependency[3]
+        skip_options = dependency[4]
         try:
-            sp.check_call(
-                ['tRNAscan-SE', '-h'],
-                stdout=sp.DEVNULL,
-                stderr=sp.DEVNULL
-            )
+            tool_output = str(sp.check_output(command, stderr=sp.STDOUT)) # stderr must be added in case the tool output is not piped into stdout
         except FileNotFoundError:
-            log.exception('tRNAscan-SE not found!')
-            sys.exit('ERROR: tRNAscan-SE not executable! Please either install tRNAscan-SE (>=v2.0.6) or skip the annotation of tRNAs via \'--skip-trna\'.')
-        except:
-            pass
-
-    # test Aragorn
-    if(cfg.skip_tmrna is False):
+            log.error('dependency not found! tool=%s', command[0])
+            sys.exit(f"ERROR: {command[0]} not found or not executable! Please make sure {command[0]} is installed and executable or skip requiring workflow steps via via '{' '.join(skip_options)}'.")
+        except sp.CalledProcessError:
+            log.error('dependency check failed! tool=%s', command[0])
+            sys.exit(f"ERROR: {command[0]} could not be executed! Please make sure {command[0]} is installed and executable or skip requiring workflow steps via via '{' '.join(skip_options)}'.")
+        version_match = re.search(version_regex, tool_output)
+        
         try:
-            sp.check_call(
-                ['aragorn', '-h'],
-                stdout=sp.DEVNULL,
-                stderr=sp.DEVNULL
-            )
-        except FileNotFoundError:
-            log.exception('aragorn not found!')
-            sys.exit('ERROR: aragorn not executable! Please either install Aragorn (>=1.2.38) or skip the annotation of tmRNAs via \'--skip-tmrna\'.')
+            if version_match is None:
+                log.error('no dependency version detected! no regex hit in dependency output: regex=%s, command=%s', version_regex, command)
+                sys.exit('ERROR: Could not detect/read %s version!', command[0])
+
+            major = version_match.group(1)
+            minor = version_match.group(2)
+            patch = version_match.group(3)
+            if major is None:
+                log.error('no dependency version detected! no regex hit in dependency output: regex=%s, command=%s', version_regex, command)
+                sys.exit('ERROR: Could not detect/read %s version!', command[0])
+            elif minor is None:
+                version_output = Version(int(major))
+            elif patch is None:
+                version_output = Version(int(major), int(minor))
+            else:
+                version_output = Version(int(major), int(minor), int(patch))
+            return version_output
         except:
-            pass
-
-    # test cmscan
-    if(cfg.skip_rrna is False or cfg.skip_ncrna is False or cfg.skip_ncrna_region is False):
-        try:
-            sp.check_call(
-                ['cmscan', '-h'],
-                stdout=sp.DEVNULL,
-                stderr=sp.DEVNULL
-            )
-        except FileNotFoundError:
-            log.exception('cmscan not found!')
-            sys.exit('ERROR: \'cmscan\' not executable! Please either install Infernal (>=v1.1.2) or skip the annotation of rRNAs, ncRNAs & ncRNA regions via \'--skip-rrna --skip-ncrna --skip-ncrna-region\'.')
-        except:
-            pass
-
-    # test prodigal
-    if(cfg.skip_cds is False):
-        try:
-            sp.check_call(
-                ['prodigal', '-v'],
-                stdout=sp.DEVNULL,
-                stderr=sp.DEVNULL
-            )
-        except FileNotFoundError:
-            log.exception('prodigal not found!')
-            sys.exit('ERROR: \'prodigal\' not executable! Please either install Prodigal (>=v2.6.3) or skip the annotation of CDSs via \'--skip-cds\'.')
-        except:
-            pass
-
-    # test hmmsearch
-    if(cfg.skip_cds is False or cfg.skip_sorf is False):
-        try:
-            sp.check_call(
-                ['hmmsearch', '-h'],
-                stdout=sp.DEVNULL,
-                stderr=sp.DEVNULL
-            )
-        except FileNotFoundError:
-            log.exception('hmmsearch not found!')
-            sys.exit('ERROR: \'hmmsearch\' not executable! Please either install HMMER (>=3.3.1) or skip the annotation of CDSs & sORFs via \'--skip-cds --skip-sorf\'.')
-        except:
-            pass
-
-    # test diamond
-    if(cfg.skip_cds is False or cfg.skip_sorf is False):
-        try:
-            sp.check_call(
-                ['diamond', '--version'],
-                stdout=sp.DEVNULL,
-                stderr=sp.DEVNULL
-            )
-        except FileNotFoundError:
-            log.exception('diamond not found!')
-            sys.exit('ERROR: \'diamond\' not executable! Please either install Diamond (>=v2.0.4) or skip the annotation of CDSs & sORFs via \'--skip-cds --skip-sorf\'.')
-        except:
-            pass
-
-    # test pilercr
-    if(cfg.skip_crispr is False):
-        try:
-            sp.check_call(
-                ['pilercr', '-options'],
-                stdout=sp.DEVNULL,
-                stderr=sp.DEVNULL
-            )
-        except FileNotFoundError:
-            log.exception('pilercr not found!')
-            sys.exit('ERROR: pilercr not executable! Please either install PILER-CR or skip the annotation of CRISPRs via \'--skip-crispr\'.')
-        except:
-            pass
+            log.error('no dependency version detected! no regex hit in dependency output: regex=%s, command=%s', version_regex, command)
+            sys.exit('ERROR: Could not detect/read %s version!', command[0])
 
 
-def create_locus_prefix(contigs):
-    """Create either genus/species or sequence MD5 hex based locus prefix.
-    Max locus name length is 37 for GenBank -> 32 + _ + 4 digits"""
-    if(cfg.genus != '' and cfg.species != ''):
-        locus_prefix = cfg.genus[:1] + cfg.species[:2]
-        return locus_prefix.upper()
+def check_version(tool, min, max):
+    """Method for checking tool versions with required version. Input: tool version, minimum and maximum version. Returns: boolean value for positive or negative check."""
+    if tool.major < min.major or tool.major > max.major:
+        return False
     else:
-        hash = hashlib.md5()
-        for contig in contigs:
-            hash.update(str.encode(contig['sequence']))
-        return hash.hexdigest()[0:5]
+        if tool.major == min.major or tool.major == max.major:
+            if tool.minor < min.minor and tool.major == min.major:
+                return False
+            elif tool.minor > max.minor and tool.major == max.major:
+                return False
+            else:
+                if tool.minor == min.minor or tool.minor == max.minor:
+                    if tool.patch < min.patch and tool.minor == min.minor:
+                        return False
+                    elif tool.patch > max.patch and tool.minor == max.minor and tool.major == max.major:
+                        return False
+                    else:
+                        return True
+                else:
+                    return True
+        else:
+            return True
+
+
+def test_dependency(dependency):
+    """Test the proper installation of the required 3rd party executable."""
+
+    version = read_tool_output(dependency)
+    check_result = check_version(version, dependency[0], dependency[1])
+    if (check_result == False):
+        log.error('wrong dependency version for %s: installed=%s, minimum=%s', dependency[3][0], version, dependency[0])
+        sys.exit(f'ERROR: Wrong {dependency[3][0]} version installed. Please, either install {dependency[3][0]} version {dependency[0]} or use {dependency[4]}!')
+    else:
+        log.info('dependency: tool=%s, version=%s', dependency[3][0], version)
+
+
+def test_dependencies():
+    """Test the proper installation of all required 3rd party executables."""
+
+    if(not cfg.skip_trna):
+        test_dependency(DEPENDENCY_TRNASCAN)
+    
+    if(not cfg.skip_tmrna):
+        test_dependency(DEPENDENCY_ARAGORN)
+    
+    if(not cfg.skip_rrna or not cfg.skip_ncrna or not cfg.skip_ncrna_region):
+        test_dependency(DEPENDENCY_CMSCAN)
+    
+    if(not cfg.skip_crispr):
+        test_dependency(DEPENDENCY_PILERCR)
+    
+    if(not cfg.skip_cds):
+        test_dependency(DEPENDENCY_PRODIGAL)
+        test_dependency(DEPENDENCY_AMRFINDERPLUS)
+
+        # test if AMRFinderPlus db is installed
+        process = sp.run(['amrfinder', '--debug'], capture_output=True)
+        if('No valid AMRFinder database found' in process.stderr.decode()):
+            log.error('AMRFinderPlus database not installed')
+            sys.exit(f"ERROR: AMRFinderPlus database not installed! Please, install AMRFinderPlus's internal database by executing: 'amrfinder -u'. This must be done only once.")
+
+    
+    if(not cfg.skip_cds or not cfg.skip_sorf):
+        test_dependency(DEPENDENCY_HMMSEARCH)
+        test_dependency(DEPENDENCY_DIAMOND)
+    
+    if(not cfg.skip_ori):
+        test_dependency(DEPENDENCY_BLASTN)
 
 
 def create_locus_tag_prefix(contigs):
@@ -187,15 +213,17 @@ def create_locus_tag_prefix(contigs):
     for contig in contigs:
         hash.update(str.encode(contig['sequence']))
     hexdigest = hash.hexdigest().upper()
-    locus_prefix = []
+    locus_prefix_chars = []
     i = 0
     while i < 6:
         c = hexdigest[i]
         if(c >= '0' and c <= '9'):
             c = chr(ord('F') + int(c) + 1) 
-        locus_prefix.append(c)
+        locus_prefix_chars.append(c)
         i += 1
-    return ''.join(locus_prefix)
+    locus_prefix = ''.join(locus_prefix_chars)
+    log.info('generated locus-tag: prefix=%s, MD5=%s', locus_prefix, hexdigest)
+    return locus_prefix
 
 
 def calc_aa_hash(seq):
@@ -212,27 +240,16 @@ def has_annotation(feature, attribute):
 
 
 def calc_genome_stats(genome, features):
-    
     genome_size = genome['size']
     log.info('genome-size=%i', genome_size)
 
     # N50
     gc_sum = 0
     n_sum = 0
-    n50 = 0
-    contig_length_sum = 0
     for contig in genome['contigs']:
         seq = contig['sequence']
         gc_sum += seq.count('G') + seq.count('C')
         n_sum += seq.count('N')
-        contig_length = len(seq)
-        contig_length_sum += contig_length
-        if(contig_length_sum >= genome_size / 2):
-            n50 = contig_length
-            break
-    genome['n50'] = n50
-    log.info('N50=%i', n50)
-
     gc_ratio = gc_sum / (genome_size - n_sum)
     genome['gc'] = gc_ratio
     log.info('GC=%0.3f', gc_ratio)
@@ -240,6 +257,17 @@ def calc_genome_stats(genome, features):
     n_ratio = n_sum / genome_size
     genome['n_ratio'] = n_ratio
     log.info('N=%0.3f', n_ratio)
+    
+    n50 = 0
+    contig_length_sum = 0
+    for contig in sorted(genome['contigs'], key=lambda x: x['length'], reverse=True):
+        contig_length = len(contig['sequence'])
+        contig_length_sum += contig_length
+        if(contig_length_sum >= genome_size / 2):
+            n50 = contig_length
+            break
+    genome['n50'] = n50
+    log.info('N50=%i', n50)
 
     contigs_by_id = {c['id']: c for c in genome['contigs']}
     coding_nts = 0
@@ -265,8 +293,11 @@ def parse_replicon_table(replicon_table_path):
     replicons = {}
     try:
         with replicon_table_path.open() as fh:
-            for line in fh:
-                (original_locus_id, new_locus_id, replicon_type, topology, name) = line.strip().split('\t')
+            dialect = csv.Sniffer().sniff(fh.read(1024), delimiters=",\t")
+            fh.seek(0)
+            reader = csv.reader(fh, dialect)
+            for row in reader:
+                (original_locus_id, new_locus_id, replicon_type, topology, name) = row
                 # TODO: add locus id checks
                 if(new_locus_id == '' or new_locus_id == ''):
                     new_locus_id = None
@@ -356,6 +387,8 @@ def qc_contigs(contigs, replicons):
                     contig['type'] = bc.REPLICON_PLASMID
             
             valid_contigs.append(contig)
+            if(len(contigs) == 1 and contig['type'] == bc.REPLICON_PLASMID and cfg.plasmid is not None):
+                contig['name'] = cfg.plasmid
             
             if(replicons):  # use user provided replicon table
                 contig_id = contig['orig_id'] if 'orig_id' in contig else contig['id']
